from itertools import chain, islice
from collections.abc import Hashable, MutableSet, Set


class FlexibleSet(MutableSet):
    """A set that uses a list (and costly identity check) for unhashable items."""

    __slots__ = ("set", "list")

    def __init__(self, iterable=None):

        self.set = set()
        self.list = []

        if iterable is not None:
            for i in iterable:
                self.add(i)

    def add(self, item):
        try:
            self.set.add(item)
        except TypeError:
            # TODO: Could try `multihash`.
            # TODO: Use `bisect` for unhashable but orderable elements
            if item not in self.list:
                self.list.append(item)

    def discard(self, item):
        try:
            self.remove(item)
        except KeyError:
            pass

    def clear(self):
        self.set.clear()
        self.list.clear()

    def pop(self):
        try:
            return self.set.pop()
        except (TypeError, KeyError):
            try:
                return self.list.pop(-1)
            except IndexError:
                raise KeyError()

    def remove(self, item):
        try:
            self.set.remove(item)
        except (TypeError, KeyError):
            try:
                self.list.remove(item)
            except ValueError:
                raise KeyError()

    def copy(self):
        res = type(self)()
        res.set = self.set.copy()
        res.list = self.list.copy()
        return res

    def __le__(self, other):
        raise NotImplementedError()

    def __ge__(self, other):
        raise NotImplementedError()

    def __ior__(self, item):
        raise NotImplementedError()

    def __iand__(self, item):
        raise NotImplementedError()

    def __ixor__(self, item):
        raise NotImplementedError()

    def __isub__(self, item):
        raise NotImplementedError()

    def __iter__(self):
        return chain(self.set, self.list)

    def __contains__(self, value):
        try:
            return value in self.set or value in self.list
        except TypeError:
            return value in self.list

    def __len__(self):
        return len(self.set) + len(self.list)

    def __eq__(self, other):
        if type(self) == type(other):
            return self.set == other.set and self.list == other.list
        elif isinstance(other, Set):
            return len(self.list) == 0 and other.issuperset(self.set)

        return NotImplemented

    def __repr__(self):
        return f"FlexibleSet([{', '.join(str(s) for s in self)}])"


def hashable(x):
    try:
        hash(x)
        return True
    except TypeError:
        return False


def dicthash(d):
    return hash(frozenset(d.items()))


def multihash(x):
    try:
        return hash(x)
    except TypeError:
        if isinstance(x, (list, tuple, set, frozenset)):
            return hash(tuple(multihash(i) for i in x))
        if type(x) is dict:
            return hash(frozenset(tuple(multihash(i) for i in x.items())))
        if type(x) is slice:
            return hash((x.start, x.stop, x.step))
        raise TypeError("Hashing not covered for " + str(x))


def unique(seq, key=lambda x: x):
    seen = set()
    for item in seq:
        try:
            k = key(item)
        except TypeError:
            # Just yield it and hope for the best, since we can't efficiently
            # check if we've seen it before.
            yield item
            continue
        if not isinstance(k, Hashable):
            # Just yield it and hope for the best, since we can't efficiently
            # check if we've seen it before.
            yield item
        elif k not in seen:
            seen.add(key(item))
            yield item


def interleave(seqs, pass_exceptions=()):
    iters = map(iter, seqs)
    while iters:
        newiters = []
        for itr in iters:
            try:
                yield next(itr)
                newiters.append(itr)
            except (StopIteration,) + tuple(pass_exceptions):
                pass
        iters = newiters


def take(n, seq):
    if n is None:
        return seq
    if n == 0:
        return tuple(seq)
    return tuple(islice(seq, 0, n))


def evalt(t):
    """Evaluate a tuple if unevaluated.

    >>> from kanren.util import evalt
    >>> add = lambda x, y: x + y
    >>> evalt((add, 2, 3))
    5
    >>> evalt(add(2, 3))
    5
    """

    if isinstance(t, tuple) and len(t) >= 1 and callable(t[0]):
        return t[0](*t[1:])
    else:
        return t


def intersection(*seqs):
    return (item for item in seqs[0] if all(item in seq for seq in seqs[1:]))


def groupsizes(total, len):
    """Construct groups of length len that add up to total.

    >>> from kanren.util import groupsizes
    >>> tuple(groupsizes(4, 2))
    ((1, 3), (2, 2), (3, 1))
    """
    if len == 1:
        yield (total,)
    else:
        for i in range(1, total - len + 1 + 1):
            for perm in groupsizes(total - i, len - 1):
                yield (i,) + perm


def pprint(g):  # pragma: no cover
    """Pretty print a tree of goals."""
    if callable(g) and hasattr(g, "__name__"):
        return g.__name__
    if isinstance(g, type):
        return g.__name__
    if isinstance(g, tuple):
        return "(" + ", ".join(map(pprint, g)) + ")"
    return str(g)


def index(tup, ind):
    """Fancy indexing with tuples."""
    return tuple(tup[i] for i in ind)



import itertools as it
from functools import partial
from .util import dicthash, interleave, take, multihash, unique, evalt
from toolz import groupby, map

from unification import reify, unify


def fail(s):
    return iter(())


def success(s):
    return iter((s,))


def eq(u, v):
    """Construct a goal stating that its arguments must unify.

    See Also
    --------
        unify
    """

    def goal_eq(s):
        result = unify(u, v, s)
        if result is not False:
            yield result

    return goal_eq


def lall(*goals):
    """Construct a logical all with goal reordering to avoid EarlyGoalErrors.

    See Also
    --------
        EarlyGoalError
        earlyorder

    >>> from kanren import lall, membero, var
    >>> x = var('x')
    >>> run(0, x, lall(membero(x, (1,2,3)), membero(x, (2,3,4))))
    (2, 3)
    """
    return (lallgreedy,) + tuple(earlyorder(*goals))


def lallgreedy(*goals):
    """Construct a logical all that greedily evaluates each goals in the order provided.

    Note that this may raise EarlyGoalError when the ordering of the
    goals is incorrect. It is faster than lall, but should be used
    with care.

    >>> from kanren import eq, run, membero, var
    >>> x, y = var('x'), var('y')
    >>> run(0, x, lallgreedy((eq, y, set([1]))), (membero, x, y))
    (1,)
    >>> run(0, x, lallgreedy((membero, x, y), (eq, y, {1})))  # doctest: +SKIP
    Traceback (most recent call last):
      ...
    kanren.core.EarlyGoalError
    """
    if not goals:
        return success
    if len(goals) == 1:
        return goals[0]

    def allgoal(s):
        g = goaleval(reify(goals[0], s))
        return unique(
            interleave(
                goaleval(reify((lallgreedy,) + tuple(goals[1:]), ss))(ss) for ss in g(s)
            ),
            key=dicthash,
        )

    return allgoal


def lallfirst(*goals):
    """Construct a logical all that runs goals one at a time.

    >>> from kanren import membero, var
    >>> x = var('x')
    >>> g = lallfirst(membero(x, (1,2,3)), membero(x, (2,3,4)))
    >>> tuple(g({}))
    ({~x: 2}, {~x: 3})
    >>> tuple(lallfirst()({}))
    ({},)
    """
    if not goals:
        return success
    if len(goals) == 1:
        return goals[0]

    def allgoal(s):
        for i, g in enumerate(goals):
            try:
                goal = goaleval(reify(g, s))
            except EarlyGoalError:
                continue
            other_goals = tuple(goals[:i] + goals[i + 1 :])
            return unique(
                interleave(
                    goaleval(reify((lallfirst,) + other_goals, ss))(ss)
                    for ss in goal(s)
                ),
                key=dicthash,
            )
        else:
            raise EarlyGoalError()

    return allgoal


def lany(*goals):
    """Construct a logical any goal.

    >>> from kanren import lany, membero, var
    >>> x = var('x')
    >>> g = lany(membero(x, (1,2,3)), membero(x, (2,3,4)))
    >>> tuple(g({}))
    ({~x: 1}, {~x: 2}, {~x: 3}, {~x: 4})
    """
    if len(goals) == 1:
        return goals[0]
    return lanyseq(goals)


def earlysafe(goal):
    """Check if a goal can be evaluated without raising an EarlyGoalError."""
    try:
        goaleval(goal)
        return True
    except EarlyGoalError:
        return False


def earlyorder(*goals):
    """Reorder goals to avoid EarlyGoalErrors.

    All goals are evaluated.  Those that raise EarlyGoalErrors are placed at
    the end in a lall

    See Also
    --------
        EarlyGoalError
    """
    if not goals:
        return ()
    groups = groupby(earlysafe, goals)
    good = groups.get(True, [])
    bad = groups.get(False, [])

    if not good:
        raise EarlyGoalError()
    elif not bad:
        return tuple(good)
    else:
        return tuple(good) + ((lall,) + tuple(bad),)


def conde(*goalseqs):
    """Construct a logical cond goal.

    Goal constructor to provides logical AND and OR

    conde((A, B, C), (D, E)) means (A and B and C) or (D and E)
    Equivalent to the (A, B, C); (D, E) syntax in Prolog.

    See Also
    --------
        lall - logical all
        lany - logical any
    """
    return (lany,) + tuple((lall,) + tuple(gs) for gs in goalseqs)


def lanyseq(goals):
    """Construct a logical any with a possibly infinite number of goals."""

    def anygoal(s):
        anygoal.goals, local_goals = it.tee(anygoal.goals)

        def f(goals):
            for goal in goals:
                try:
                    yield goaleval(reify(goal, s))(s)
                except EarlyGoalError:
                    pass

        return unique(
            interleave(f(local_goals), pass_exceptions=[EarlyGoalError]), key=dicthash
        )

    anygoal.goals = goals

    return anygoal


def condeseq(goalseqs):
    """Construct a goal like conde, but support generic, possibly infinite iterators of goals."""
    return (lanyseq, ((lall,) + tuple(gs) for gs in goalseqs))


def everyg(predicate, coll):
    """Assert that a predicate applies to all elements of a collection."""
    return (lall,) + tuple((predicate, x) for x in coll)


def run(n, x, *goals):
    """Run a logic program and obtain n solutions that satisfy the given goals.

    >>> from kanren import run, var, eq
    >>> x = var()
    >>> run(1, x, eq(x, 1))
    (1,)

    Parameters
    ----------
    n: int
        The number of desired solutions (see `take`). `n=0` returns a tuple
        with all results and `n=None` returns a lazy sequence of all results.
    x: object
        The form to reify and output.  Usually contains logic variables used in
        the given goals.
    goals: Callables
        A sequence of goals that must be true in logical conjunction
        (i.e. `lall`).
    """
    results = map(partial(reify, x), goaleval(lall(*goals))({}))
    return take(n, unique(results, key=multihash))


class EarlyGoalError(Exception):
    """An exception indicating that a goal has been constructed prematurely.

    Consider the following case

    >>> from kanren import run, eq, membero, var
    >>> x, coll = var(), var()
    >>> run(0, x, (membero, x, coll), (eq, coll, (1, 2, 3))) # doctest: +SKIP

    The first goal, membero, iterates over an infinite sequence of all possible
    collections.  This is unproductive.  Rather than proceed, membero raises an
    EarlyGoalError, stating that this goal has been called early.

    The goal constructor lall Logical-All-Early will reorder such goals to
    the end so that the call becomes

    >>> run(0, x, (eq, coll, (1, 2, 3)), (membero, x, coll)) # doctest: +SKIP

    In this case coll is first unified to ``(1, 2, 3)`` then x iterates over
    all elements of coll, 1, then 2, then 3.

    See Also
    --------
        lall
        earlyorder
    """


def find_fixed_point(f, arg):
    """Repeatedly calls f until a fixed point is reached.

    This may not terminate, but should if you apply some eventually-idempotent
    simplification operation like evalt.
    """
    last, cur = object(), arg
    while last != cur:
        last = cur
        cur = f(cur)
    return cur


def goaleval(goal):
    """Expand and then evaluate a goal.

    See Also
    --------
       goalexpand
    """
    if callable(goal):  # goal is already a function like eq(x, 1)
        return goal
    if isinstance(goal, tuple):  # goal is not yet evaluated like (eq, x, 1)
        return find_fixed_point(evalt, goal)
    raise TypeError("Expected either function or tuple")



import operator
import collections

from itertools import permutations
from collections.abc import Sequence

from cons import cons
from cons.core import ConsNull
from unification import isvar, var, reify, unify

from .core import (
    eq,
    EarlyGoalError,
    conde,
    condeseq,
    lany,
    lallgreedy,
    lall,
    fail,
    success,
)
from .util import unique


def heado(head, coll):
    """Construct a goal stating that head is the head of coll.

    See Also
    --------
        tailo
        conso
    """
    return eq(cons(head, var()), coll)


def tailo(tail, coll):
    """Construct a goal stating that tail is the tail of coll.

    See Also
    --------
        heado
        conso
    """
    return eq(cons(var(), tail), coll)


def conso(h, t, l):
    """Construct a goal stating that cons h + t == l."""
    return eq(cons(h, t), l)


def nullo(l):
    """Create a goal asserting that a term is a "Lisp-like" null.

    For un-unified logic variables, it unifies with an empty list.
    """

    def _nullo(s):
        l_rf = reify(l, s)
        if isvar(l_rf):
            yield unify(l_rf, [], s)
        elif isinstance(l_rf, ConsNull):
            yield s

    return _nullo


def itero(l):
    """Construct a goal asserting that a term is an iterable type.

    This is a generic version of the standard `listo` that accounts for
    different iterable types supported by `cons` in Python.

    See `nullo`
    """
    c, d = var(), var()
    return (conde, [(nullo, l), success], [(conso, c, d, l), (itero, d)])


def permuteq(a, b, eq2=eq):
    """Construct a goal asserting equality under permutation.

    For example, (1, 2, 2) equates to (2, 1, 2) under permutation
    >>> from kanren import var, run, permuteq
    >>> x = var()
    >>> run(0, x, permuteq(x, (1, 2)))
    ((1, 2), (2, 1))

    >>> run(0, x, permuteq((2, 1, x), (2, 1, 2)))
    (2,)
    """
    if isinstance(a, Sequence) and isinstance(b, Sequence):
        if len(a) != len(b):
            return fail
        elif collections.Counter(a) == collections.Counter(b):
            return success
        else:
            c, d = list(a), list(b)
            for x in list(c):
                # TODO: This is quadratic in the number items in the sequence.
                # Need something like a multiset. Maybe use
                # collections.Counter?
                try:
                    d.remove(x)
                    c.remove(x)
                except ValueError:
                    pass

            if len(c) == 1:
                return (eq2, c[0], d[0])
            return condeseq(
                ((eq2, x, d[0]), (permuteq, c[0:i] + c[i + 1 :], d[1:], eq2))
                for i, x in enumerate(c)
            )
    elif not (isinstance(a, Sequence) or isinstance(b, Sequence)):
        raise ValueError(
            "Neither a nor b is a Sequence: {}, {}".format(type(a), type(b))
        )

    if isvar(a) and isvar(b):
        raise EarlyGoalError()

    if isvar(a) or isvar(b):
        if isinstance(b, Sequence):
            c, d = a, b
        elif isinstance(a, Sequence):
            c, d = b, a

        return (condeseq, ([eq(c, perm)] for perm in unique(permutations(d, len(d)))))


def seteq(a, b, eq2=eq):
    """Construct a goal asserting set equality.

    For example (1, 2, 3) set equates to (2, 1, 3)

    >>> from kanren import var, run, seteq
    >>> x = var()
    >>> run(0, x, seteq(x, (1, 2)))
    ((1, 2), (2, 1))

    >>> run(0, x, seteq((2, 1, x), (3, 1, 2)))
    (3,)
    """
    ts = lambda x: tuple(set(x))
    if not isvar(a) and not isvar(b):
        return permuteq(ts(a), ts(b), eq2)
    elif not isvar(a):
        return permuteq(ts(a), b, eq2)
    else:  # not isvar(b)
        return permuteq(a, ts(b), eq2)


def goalify(func, name=None):
    """Convert Python function into kanren goal.

    >>> from kanren import run, goalify, var, membero
    >>> typo = goalify(type)
    >>> x = var('x')
    >>> run(0, x, membero(x, (1, 'cat', 'hat', 2)), (typo, x, str))
    ('cat', 'hat')

    Goals go both ways.  Here are all of the types in the collection

    >>> typ = var('typ')
    >>> results = run(0, typ, membero(x, (1, 'cat', 'hat', 2)), (typo, x, typ))
    >>> print([result.__name__ for result in results])
    ['int', 'str']
    """

    def funco(inputs, out):
        if isvar(inputs):
            raise EarlyGoalError()
        else:
            if isinstance(inputs, (tuple, list)):
                if any(map(isvar, inputs)):
                    raise EarlyGoalError()
                return (eq, func(*inputs), out)
            else:
                return (eq, func(inputs), out)

    name = name or (func.__name__ + "o")
    funco.__name__ = name

    return funco


def membero(x, coll):
    """Construct a goal stating that x is an item of coll."""
    if not isvar(coll):
        return (lany,) + tuple((eq, x, item) for item in coll)
    raise EarlyGoalError()


typo = goalify(type, name="typo")
isinstanceo = goalify(isinstance, name="isinstanceo")
not_equalo = goalify(operator.ne, name="not_equalo")


def appendo(l, s, ls, base_type=tuple):
    """Construct a goal stating ls = l + s.

    See Byrd thesis pg. 247
    https://scholarworks.iu.edu/dspace/bitstream/handle/2022/8777/Byrd_indiana_0093A_10344.pdf

    Parameters
    ----------
    base_type: type
        The empty collection type to use when all terms are logic variables.
    """
    if all(map(isvar, (l, s, ls))):
        raise EarlyGoalError()
    a, d, res = [var() for i in range(3)]
    return (
        lany,
        (lallgreedy, (eq, l, base_type()), (eq, s, ls)),
        (lall, (conso, a, d, l), (conso, a, res, ls), (appendo, d, s, res)),
    )



import operator

from unification import isvar

from .core import eq, EarlyGoalError, lany


def gt(x, y):
    """Construct a goal stating x > y."""
    if not isvar(x) and not isvar(y):
        return eq(x > y, True)
    else:
        raise EarlyGoalError()


def lt(x, y):
    """Construct a goal stating x > y."""
    if not isvar(x) and not isvar(y):
        return eq(x < y, True)
    else:
        raise EarlyGoalError()


def lor(*goalconsts):
    """Construct a goal representing a logical OR for goal constructors.

    >>> from kanren.arith import lor, eq, gt
    >>> gte = lor(eq, gt)  # greater than or equal to is `eq or gt`
    """

    def goal(*args):
        return lany(*[gc(*args) for gc in goalconsts])

    return goal


gte = lor(gt, eq)
lte = lor(lt, eq)


def binop(op, revop=None):
    """Transform binary operator into goal.

    >>> from kanren.arith import binop
    >>> import operator
    >>> add = binop(operator.add, operator.sub)

    >>> from kanren import var, run
    >>> x = var('x')
    >>> next(add(1, 2, x)({}))
    {~x: 3}
    """

    def goal(x, y, z):
        if not isvar(x) and not isvar(y):
            return eq(op(x, y), z)
        if not isvar(y) and not isvar(z) and revop:
            return eq(x, revop(z, y))
        if not isvar(x) and not isvar(z) and revop:
            return eq(y, revop(z, x))
        raise EarlyGoalError()

    goal.__name__ = op.__name__
    return goal


add = binop(operator.add, operator.sub)
add.__doc__ = """ x + y == z """
mul = binop(operator.mul, operator.truediv)
mul.__doc__ = """ x * y == z """
mod = binop(operator.mod)
mod.__doc__ = """ x % y == z """


def sub(x, y, z):
    """Construct a goal stating x - y == z."""
    return add(y, z, x)


def div(x, y, z):
    """Construct a goal stating x / y == z."""
    return mul(z, y, x)



# flake8: noqa
"""kanren is a Python library for logic and relational programming."""
from __future__ import absolute_import

from unification import unify, reify, unifiable, var, isvar, vars, variables, Var

from .core import run, eq, conde, lall, lany
from .goals import seteq, permuteq, goalify, membero
from .facts import Relation, fact, facts
from .term import arguments, operator, term, unifiable_with_term

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



from unification import unify, reify
from unification.core import _unify, _reify

from .dispatch import dispatch


@dispatch((tuple, list))
def arguments(seq):
    return seq[1:]


@dispatch((tuple, list))
def operator(seq):
    return seq[0]


@dispatch(object, (tuple, list))
def term(op, args):
    return (op,) + tuple(args)


def unifiable_with_term(cls):
    _reify.add((cls, dict), reify_term)
    _unify.add((cls, cls, dict), unify_term)
    return cls


def reify_term(obj, s):
    op, args = operator(obj), arguments(obj)
    op = reify(op, s)
    args = reify(args, s)
    new = term(op, args)
    return new


def unify_term(u, v, s):
    u_op, u_args = operator(u), arguments(u)
    v_op, v_args = operator(v), arguments(v)
    s = unify(u_op, v_op, s)
    if s is not False:
        s = unify(u_args, v_args, s)
    return s



"""Functions for associative and commutative unification.

This module provides goals for associative and commutative unification.  It
accomplishes this through naively trying all possibilities.  This was built to
be used in the computer algebra systems SymPy and Theano.

>>> from kanren import run, var, fact
>>> from kanren.assoccomm import eq_assoccomm as eq
>>> from kanren.assoccomm import commutative, associative

>>> # Define some dummy Ops
>>> add = 'add'
>>> mul = 'mul'

>>> # Declare that these ops are commutative using the facts system
>>> fact(commutative, mul)
>>> fact(commutative, add)
>>> fact(associative, mul)
>>> fact(associative, add)

>>> # Define some wild variables
>>> x, y = var('x'), var('y')

>>> # Two expressions to match
>>> pattern = (mul, (add, 1, x), y)                # (1 + x) * y
>>> expr    = (mul, 2, (add, 3, 1))                # 2 * (3 + 1)

>>> print(run(0, (x,y), eq(pattern, expr)))
((3, 2),)
"""

from unification.utils import transitive_get as walk
from unification import isvar, var


from . import core
from .core import (
    unify,
    conde,
    eq,
    fail,
    lallgreedy,
    EarlyGoalError,
    condeseq,
    goaleval,
)
from .goals import permuteq
from .facts import Relation
from .util import groupsizes, index
from .term import term, arguments, operator

associative = Relation("associative")
commutative = Relation("commutative")


def assocunify(u, v, s, eq=core.eq, n=None):
    """Perform associative unification.

    See Also
    --------
        eq_assoccomm
    """
    uop, uargs = op_args(u)
    vop, vargs = op_args(v)

    if not uop and not vop:
        res = unify(u, v, s)
        if res is not False:
            return (res,)  # TODO: iterate through all possibilities

    if uop and vop:
        s = unify(uop, vop, s)
        if s is False:
            return ().__iter__()
        op = walk(uop, s)

        sm, lg = (uargs, vargs) if len(uargs) <= len(vargs) else (vargs, uargs)
        ops = assocsized(op, lg, len(sm))
        goal = condeseq([(eq, a, b) for a, b, in zip(sm, lg2)] for lg2 in ops)
        return goaleval(goal)(s)

    if uop:
        op, tail = uop, uargs
        b = v
    if vop:
        op, tail = vop, vargs
        b = u

    ns = [n] if n else range(2, len(tail) + 1)
    knowns = (build(op, x) for n in ns for x in assocsized(op, tail, n))

    goal = condeseq([(core.eq, b, k)] for k in knowns)
    return goaleval(goal)(s)


def assocsized(op, tail, n):
    """Produce all associative combinations of x in n groups."""
    gsizess = groupsizes(len(tail), n)
    partitions = (groupsizes_to_partition(*gsizes) for gsizes in gsizess)
    return (makeops(op, partition(tail, part)) for part in partitions)


def makeops(op, lists):
    """Construct operations from an op and parition lists.

    >>> from kanren.assoccomm import makeops
    >>> makeops('add', [(1, 2), (3, 4, 5)])
    (('add', 1, 2), ('add', 3, 4, 5))
    """
    return tuple(l[0] if len(l) == 1 else build(op, l) for l in lists)


def partition(tup, part):
    """Partition a tuple.

    >>> from kanren.assoccomm import partition
    >>> partition("abcde", [[0,1], [4,3,2]])
    [('a', 'b'), ('e', 'd', 'c')]
    """
    return [index(tup, ind) for ind in part]


def groupsizes_to_partition(*gsizes):
    """Create a list of ranges from their sizes.

    >>> from kanren.assoccomm import groupsizes_to_partition
    >>> groupsizes_to_partition(2, 3)
    [[0, 1], [2, 3, 4]]
    """
    idx = 0
    part = []
    for gs in gsizes:
        l_ = []
        for i in range(gs):
            l_.append(idx)
            idx += 1
        part.append(l_)
    return part


def eq_assoc(u, v, eq=core.eq, n=None):
    """Create a goal for associative equality.

    >>> from kanren import run, var, fact
    >>> from kanren.assoccomm import eq_assoc as eq

    >>> fact(commutative, 'add')    # declare that 'add' is commutative
    >>> fact(associative, 'add')    # declare that 'add' is associative

    >>> x = var()
    >>> run(0, x, eq(('add', 1, 2, 3), ('add', 1, x)))
    (('add', 2, 3),)
    """
    uop, _ = op_args(u)
    vop, _ = op_args(v)
    if uop and vop:
        return conde(
            [(core.eq, u, v)],
            [(eq, uop, vop), (associative, uop), lambda s: assocunify(u, v, s, eq, n)],
        )

    if uop or vop:

        if vop:
            uop, vop = vop, uop
            v, u = u, v
        return conde(
            [(core.eq, u, v)],
            [(associative, uop), lambda s: assocunify(u, v, s, eq, n)],
        )

    return (core.eq, u, v)


def eq_comm(u, v, eq=None):
    """Create a goal for commutative equality.

    >>> from kanren import run, var, fact
    >>> from kanren.assoccomm import eq_comm as eq
    >>> from kanren.assoccomm import commutative, associative

    >>> fact(commutative, 'add')    # declare that 'add' is commutative
    >>> fact(associative, 'add')    # declare that 'add' is associative

    >>> x = var()
    >>> run(0, x, eq(('add', 1, 2, 3), ('add', 2, x, 1)))
    (3,)
    """
    eq = eq or eq_comm
    vtail = var()
    if isvar(u) and isvar(v):
        return (core.eq, u, v)
    uop, uargs = op_args(u)
    vop, vargs = op_args(v)
    if not uop and not vop:
        return (core.eq, u, v)
    if vop and not uop:
        uop, uargs = vop, vargs
        v, u = u, v
    return (
        conde,
        ((core.eq, u, v),),
        ((commutative, uop), (buildo, uop, vtail, v), (permuteq, uargs, vtail, eq)),
    )


def buildo(op, args, obj):
    """Construct a goal that relates an object to its op and args.

    For example, in `add(1,2,3)`, `add` is the op and `(1,2,3)` are the args.

    Checks op_regsitry for functions to define op/arg relationships
    """
    if not isvar(obj):
        oop, oargs = op_args(obj)
        # TODO: Is greedy correct?
        return lallgreedy((eq, op, oop), (eq, args, oargs))
    else:
        try:
            return eq(obj, build(op, args))
        except TypeError:
            raise EarlyGoalError()
    raise EarlyGoalError()


def build(op, args):
    try:
        return term(op, args)
    except NotImplementedError:
        raise EarlyGoalError()


def op_args(x):
    """Break apart x into an operation and tuple of args."""
    if isvar(x):
        return None, None
    try:
        return operator(x), arguments(x)
    except NotImplementedError:
        return None, None


def eq_assoccomm(u, v):
    """Construct a goal for associative and commutative eq.

    Works like logic.core.eq but supports associative/commutative expr trees

    tree-format:  (op, *args)
    example:      (add, 1, 2, 3)

    State that operations are associative or commutative with relations

    >>> from kanren.assoccomm import eq_assoccomm as eq
    >>> from kanren.assoccomm import commutative, associative
    >>> from kanren import fact, run, var

    >>> fact(commutative, 'add')    # declare that 'add' is commutative
    >>> fact(associative, 'add')    # declare that 'add' is associative

    >>> x = var()
    >>> e1 = ('add', 1, 2, 3)
    >>> e2 = ('add', 1, x)
    >>> run(0, x, eq(e1, e2))
    (('add', 2, 3), ('add', 3, 2))
    """
    uop, uargs = op_args(u)
    vop, vargs = op_args(v)

    if uop and not vop and not isvar(v):
        return fail
    if vop and not uop and not isvar(u):
        return fail
    if uop and vop and uop != vop:
        return fail
    if uop and not (uop,) in associative.facts:
        return (eq, u, v)
    if vop and not (vop,) in associative.facts:
        return (eq, u, v)

    if uop and vop:
        u, v = (u, v) if len(uargs) >= len(vargs) else (v, u)
        n = min(map(len, (uargs, vargs)))  # length of shorter tail
    else:
        n = None
    if vop and not uop:
        u, v = v, u
    w = var()
    # TODO: Is greedy correct?
    return (
        lallgreedy,
        (eq_assoc, u, w, eq_assoccomm, n),
        (eq_comm, v, w, eq_assoccomm),
    )
